// global vars

var sel;
var form;
var splitTicks;

// Objects

function ScoreItem(itype, item) {// chord or rest
    if (itype == "chord" || itype == "rest") {
        this.siType = itype;
        this.item = item;
        this.exactLen = item.tickLen;
    }
    else {
        this.siType = "undefined";
        this.item = null;
        this.exactLen = 0;
    }

    this.getLen = function siGetLen() {
        return this.exactLen;
    }

    this.setLen = function siGetLen(len) {
        this.exactLen = len;
        this.item.tickLen = len;
    }

}

function ScoreItemList() {
    this.entity = []
    this.length = 0;

    this.add = function silAdd(si) {
        this.length++;
        this.entity[this.length - 1] = si;
    }

    this.del = function silDel(pos) {
        for (i = pos; i < this.length -1; i++) {
            this.entity[i] = this.entity[i + 1];
        }
        this.length--;
    }

    this.insertList = function silInsertList(pos, sil) {
        for (i = this.length - 1; i >= pos; i--) {
            this.entity[i + sil.length] = this.entity[i];
        }
        this.length += sil.length;
        for (i = 0; i < sil.length; i++) {
            this.entity[pos + i] = sil.entity[i];
        }
    }

    this.insertItem = function silInsertList(pos, si) {
        for (i = this.length - 1; i >= pos; i--) {
            this.entity[i + 1] = this.entity[i];
        }
        this.length += 1;
        this.entity[pos] = si;
    }

    this.getLen = function silGetLen(pos) {
        return this.entity[pos].getLen();
    }

    this.setLen = function silGetLen(pos, len) {
        this.entity[pos].setLen(len);
    }

    this.copy = function silCopy(pos) {
        var si = new ScoreItem();
        si.siType = this.entity[pos].siType;
        if (si.siType == "chord") {
            si.item = copyChord(this.entity[pos].item);
        }
        if (si.siType == "rest") {
            si.item = copyRest(this.entity[pos].item);
        }
        si.exactLen = this.getLen(pos);
        si.item.tickLen = si.exactLen;
        return si;
    }

    this.fitIntoMeasures = function silFitIntoMeasures(mTicks) {
        var nextTick = 0;
        var iMeasure = 1;
        var iScoreItem = 0;
        while (iScoreItem < this.length) {
            var nextLen = this.getLen(iScoreItem);
            nextTick += nextLen;
            if (nextTick == mTicks[iMeasure]) {
                iMeasure++;
            }
            else if (nextTick > mTicks[iMeasure]) {
                var len2 = nextTick - mTicks[iMeasure];
                var len1 = nextLen - len2;

                var aTicks1 = splitNote(len1, true);
                var aTicks2 = splitNote(len2, false);

                var sil = new ScoreItemList();
                for (iSplit = 0; iSplit < aTicks1.length; iSplit++){
                    sil.add(this.copy(iScoreItem));
                    sil.setLen(iSplit, aTicks1[iSplit]);
                }
                for (iSplit = 0; iSplit < aTicks2.length; iSplit++){
                    sil.add(this.copy(iScoreItem));
                    sil.setLen(aTicks1.length + iSplit, aTicks2[iSplit]);
                }
                this.del(iScoreItem);
                this.insertList(iScoreItem, sil);
                nextTick = mTicks[iMeasure];
                iMeasure++;
            }
            iScoreItem += 1;
        }
    }
}

// functions

function setSplitTicks (sMode) {// accepted note lengths
    splitTicks = [];
    switch (sMode) {
    case "undotted":
        splitTicks = [1920, 960, 480, 240, 120, 60, 30];
        break;
    case "dotted":
        splitTicks = [1920, 1440, 960, 720, 480, 360, 240, 180, 120, 90, 60, 30];
        break;
    case "doubledotted":
        splitTicks = [1920, 1680, 1440, 960, 840, 720, 480, 420, 360, 240, 210, 180, 120, 105, 90, 60, 30];
        break;    
    }
}

function refresh() {
    var sc = new Score();
    sc.close();
}


function getScoreItem(cur) {// gets chord/rest at cursor position
    var si = null;
    if (cur.isChord()) {
        si = new ScoreItem("chord", copyChord(cur.chord()));
    }
    else if (cur.isRest()) {
        si = new ScoreItem("rest", copyRest(cur.rest()));
    }
    return si;
}


function getMeasureTicks(cur, endTick, relative) {// returns array with ticks of all measures between cur.selectionStart and endTick
    var aTicks =[];
    var iMeasure = 0;
    var startTick = cur.tick();
    while (cur.tick() < endTick) {
        aTicks[iMeasure] = cur.tick();
        if (relative) {
            aTicks[iMeasure] -= startTick;
        }
        cur.nextMeasure();
        iMeasure++;
    }
    aTicks[iMeasure] = endTick;
    return aTicks;
}

function Selection(sScore) {//returns some Data of the current selection
    this.score = sScore;
    this.cursor = new Cursor(sScore);
    this.cursor.goToSelectionStart();
    if (this.cursor.eos()) {
        this.exists = false;
        this.staffStart = 0;
        this.tickStart = 0;
        this.staffEnd = this.score.staves - 1;
        this.tickEnd = this.cursor.tick();
    }
    else {
        this.exists = true;
        this.staffStart = this.cursor.staff;
        this.tickStart = this.cursor.tick();
        this.cursor.goToSelectionEnd();
        this.staffEnd = this.cursor.staff - 1;
        this.tickEnd = this.cursor.tick();
    }
    this.cursor.goToSelectionStart();
    this.measureTicks = getMeasureTicks(this.cursor, this.tickEnd, false);
}

function getVoice() {// retrieve input voice
    var selVc = 0;
    if (form.groupVoice.buttVoice1.checked) { selVc = 0; }
    if (form.groupVoice.buttVoice2.checked) { selVc = 1; }
    if (form.groupVoice.buttVoice3.checked) { selVc = 2; }
    if (form.groupVoice.buttVoice4.checked) { selVc = 3; }
    return selVc;
}

function setSplitting() {// retrieve input splitting
    if (form.groupSplitting.buttUndotted.checked) { setSplitTicks("undotted"); }
    if (form.groupSplitting.buttDotted.checked) { setSplitTicks("dotted"); }
    if (form.groupSplitting.buttDoubleDotted.checked) { setSplitTicks("doubledotted"); }
}

function loadForm(formfile) {
    var loader = new QUiLoader(null);
    var file   = new QFile(pluginPath + "/" + formfile);
    file.open(QIODevice.OpenMode(QIODevice.ReadOnly, QIODevice.Text));
    return loader.load(file, null);
}

function goTo(cur, tick) {// move cur forward to tick
    while (!cur.eos() && cur.tick() < tick) {
        cur.next();
    }
}

function copyRest(srcRest) {// makes a duplicate of a rest
    var dstRest = new Rest();
    dstRest.tickLen = srcRest.tickLen;
    return dstRest;
}

function copyNote(srcNote) {// makes a duplicate of a note
    var dstNote = new Note();
    dstNote.boundingRect = srcNote.boundingRect;
    dstNote.color = srcNote.color;
    dstNote.noteHead = srcNote.noteHead;
    dstNote.pitch = srcNote.pitch;
    dstNote.tied = srcNote.tied;
    dstNote.tpc = srcNote.tpc;
    dstNote.tuning = srcNote.tuning;
    dstNote.visible = srcNote.visible;
//    dstNote.userAccidental = srcNote.userAccidental;      hangs up
    dstNote.velocity = srcNote.velocity;
    return dstNote;
}

function copyChord(srcChord) {// makes a duplicate of a chord
    var dstChord = new Chord();
//    dstChord.noStem = srcChord.noStem;        hangs up
    dstChord.tickLen = srcChord.tickLen;
    dstChord.siType = srcChord.siType;
    for (iNote = 0; iNote < srcChord.notes; iNote++) {
        dstChord.addNote(copyNote(srcChord.note(iNote)));
    }
    return dstChord;
}

function splitNote(ticks, ascending) {//splits any value into reasonable quantums of ticks
    var aTicks = [];
    var iSplit = 0;
    while (ticks >= 30 && iSplit < splitTicks.length) {
        while (iSplit < splitTicks.length - 1 && splitTicks[iSplit] > ticks) {
            iSplit++;
        }
        aTicks.push(splitTicks[iSplit]);
        ticks -= splitTicks[iSplit];
    }
    if (ascending) {
        aTicks.reverse();
    }
    return aTicks;
}

// main functions copy/cut/insert

function copy() {
    copycut(false);
}

function cut() {
    copycut(true);
}

function copycut(cutMode) {
    var selVoice = getVoice();
    setSplitting();
    var cbScore = new Score();
    cbScore.load(pluginPath+"/emptyclipboard.mscz")
    var cbCursor = new Cursor(cbScore);
    sel.cursor.goToSelectionStart();
    sel.cursor.voice = selVoice;
    cbCursor.rewind();
// write selection to clipboard
    while (!sel.cursor.eos() &&  sel.cursor.tick() < sel.tickEnd) {
        if (sel.cursor.isChord()) {
            cbCursor.add(copyChord(sel.cursor.chord()));
            cbCursor.next();
        }
        else if (sel.cursor.isRest()) {
            cbCursor.add(copyRest(sel.cursor.rest()));
            cbCursor.next();
        }
        sel.cursor.next();
    }
// endmarker and close
    note = new Note();
    note.pitch = 127;
    chord = new Chord();
    chord.addNote(note);
    chord.tickLen = 30;
    cbCursor.add(chord);
    cbScore.save(pluginPath+"/clipboard","mscz");
    cbScore.close();// evtl erst am schluss
    if (cutMode) {
// replace selection with rests
        sel.cursor.goToSelectionStart();
        sel.cursor.voice = selVoice;
// get the fitting ticklength for all measures
        var aTick = sel.measureTicks;
        //aTick = getMeasureTicks(sel.cursor, sel.endTick, false);
        var tick = aTick[1] - aTick[0];
        var sTick = splitNote(tick, true);
        sTick.unshift(aTick[0]);
        for (var i = 1 ; i < sTick.length - 1; i++) {
            sTick[i] += sTick[i - 1];
        }
        sTick.pop();
        aTick.shift();
        for (i = sTick.length - 1; i >= 0; i--) {
            aTick.unshift(sTick[i]);
        }
        tick = aTick[aTick.length - 1] - aTick[aTick.length - 2];
        sTick = splitNote(tick, false);
        sTick.unshift(aTick[aTick.length - 2]);
        for (var i = 1; i < sTick.length; i++) {
            sTick[i] += sTick[i - 1];
        }
        sTick.shift();
        aTick.pop();
        for (i = 0; i < sTick.length; i++) {
            aTick.push(sTick[i]);
        }
// write rests
        sel.cursor.goToSelectionStart();
        sel.cursor.voice = selVoice;
        for (i=0; i < (aTick.length - 1); i++) {
            var rest = new Rest();
            rest.tickLen = aTick[i+1] - aTick[i];
            var nextTick = sel.cursor.tick() + rest.tickLen;
            sel.cursor.add(rest);
            goTo(sel.cursor, nextTick);
        }

    }
    if (cutMode) {
        refresh();
    }
    form.close();
}



function insert() {
    var cbCursor;

        function endmarker() {
            var em = false;
            if (cbCursor.isChord()) {
                chord = cbCursor.chord();
                if (chord.note(0).pitch == 127) {
                    em = true;
                }
            }
            return em;
        }

// read from clipboard
    var aNotes = new ScoreItemList();
    var selVoice = getVoice();
    setSplitting();
    var cbScore = new Score();
    cbScore.load(pluginPath+"/clipboard.mscz")
    cbCursor = new Cursor(cbScore);
    cbCursor.rewind();
    while (!cbCursor.eos() && !endmarker()) {
        var si = getScoreItem(cbCursor);
        if (si != null) {
            aNotes.add(si);
        }
        cbCursor.next();
    }
    var cbTicks = cbCursor.tick();
// write to array aNotes
    sel.cursor.goToSelectionStart();
    aMTicks = getMeasureTicks(sel.cursor, sel.tickStart + cbTicks, true);
    aNotes.fitIntoMeasures(aMTicks);
// insert at selection.start
    var iNotes = 0;
    sel.cursor.goToSelectionStart();
    sel.cursor.voice = selVoice;
    while (!sel.cursor.eos() && iNotes < aNotes.length) {
        var nextTick = sel.cursor.tick() + aNotes.getLen(iNotes);
        sel.cursor.add(aNotes.entity[iNotes].item);
        goTo(sel.cursor, nextTick);
        iNotes++;
    }
    cbScore.close();
    refresh();
    form.close();
}

// standard functions

function init() {
}


function run() {
    if (typeof curScore === 'undefined') {
        return;
    }
    sel = new Selection(curScore);
    if (sel.exists) {
        form = loadForm("handlevoiceform.ui");
        form.groupTask.buttCopy.clicked.connect(copy);
        form.groupTask.buttCut.clicked.connect(cut);
        form.groupTask.buttInsert.clicked.connect(insert);
        setSplitTicks("dotted");
        form.show();
    }
    else {
        form = loadForm("errorform.ui");
        form.buttAccept.clicked.connect(form.close);
    }
    form.show();
}

var mscorePlugin = {
    menu: 'Plugins.Voices.HandleVoice',
    init: init,
    run: run
}
mscorePlugin
